"use strict";
var cli = require("./cli.js"), _package = require("./package.js"), eslint = require("eslint"), path = require("path"), outdent = require("outdent"), chalk = require("chalk"), childProcess = require("child_process"), npmRunPath = require("npm-run-path"), pkgUtils = require("@sanity/pkg-utils");
function _interopDefaultCompat(e) {
  return e && typeof e == "object" && "default" in e ? e : { default: e };
}
var path__default = /* @__PURE__ */ _interopDefaultCompat(path), outdent__default = /* @__PURE__ */ _interopDefaultCompat(outdent), chalk__default = /* @__PURE__ */ _interopDefaultCompat(chalk), childProcess__default = /* @__PURE__ */ _interopDefaultCompat(childProcess), npmRunPath__default = /* @__PURE__ */ _interopDefaultCompat(npmRunPath);
const removedImportSuffix = `imports where removed in Sanity v3. Please refer to the migration guide: ${cli.urls.migrationGuideStudio}, or new API-reference docs: ${cli.urls.refDocs}`;
async function validateImports({ basePath }) {
  cli.log.debug("Running ESLint with Sanity Studio import hints...");
  const eslint$1 = new eslint.ESLint({
    cwd: basePath,
    overrideConfig: {
      ignorePatterns: ["node_modules"],
      rules: {
        "no-restricted-imports": [
          "error",
          {
            patterns: [
              ..._package.mergedPackages.map((packageName) => ({
                group: [`${packageName}*`],
                message: `Use sanity instead of ${packageName}.`
              })),
              {
                group: ["config:*"],
                message: `config: imports are no longer supported. Please see the new plugin API for alternatives: ${cli.urls.migrationGuideStudio}`
              },
              {
                group: ["part:*"],
                message: `part: ${removedImportSuffix}`
              },
              {
                group: ["all:part:*"],
                message: `all:part: ${removedImportSuffix}`
              },
              {
                group: ["sanity:*"],
                message: `sanity: ${removedImportSuffix}`
              }
            ]
          }
        ]
      }
    }
  });
  try {
    const onlyImportErrors = (await eslint$1.lintFiles([path__default.default.join(basePath, "**/*.{js,jsx,ts,tsx}")])).map((r) => {
      const limitErrors = r.messages.filter((m) => m.ruleId === "no-restricted-imports");
      return {
        ...r,
        messages: limitErrors,
        errorCount: limitErrors.length
      };
    }).filter((r) => r.errorCount);
    if (onlyImportErrors.length) {
      const resultText = await (await eslint$1.loadFormatter("stylish")).format(onlyImportErrors), addtionalInfo = outdent__default.default`
        ESLint detected Studio V2 imports that are no longer available.
        It is recommended configure @sanity/eslint-config-no-v2-imports for ESLint.

        Run:
        npm install --save-dev @sanity/eslint-config-no-v2-imports

        In .eslintrc add:
        "extends": ["@sanity/no-v2-imports"]

        This way, V2-imports can be identified directly in the IDE, or using eslint CLI.
        For more, see ${cli.urls.linterPackage}

        If the plugin package does not use eslint, disable this check.
    `;
      return [resultText + addtionalInfo];
    }
  } catch (e) {
    return cli.log.error("Failed to run eslint check", e), [
      outdent__default.default`
        Failed to run ESLint. Is ESLint configured?

        If the package does not use eslint, disable this check.
    `
    ];
  }
  return [];
}
function parseCommand(commandString) {
  const commandAndArg = commandString.replace(/ +/g, " ").split(" ");
  return {
    command: commandAndArg[0],
    args: commandAndArg.length > 1 ? commandAndArg.slice(1) : []
  };
}
async function runCommand(commandString) {
  cli.log.info(`Running command: ${commandString}`);
  const { command, args } = parseCommand(commandString);
  let options = { stdio: "inherit", env: npmRunPath__default.default.env() };
  return options = process.platform === "win32" ? { ...options, shell: !0 } : options, new Promise((resolve, reject) => {
    childProcess__default.default.spawn(command, args, options).on("error", reject).on("close", (exitCode) => {
      resolve({ code: exitCode ?? 0 });
    });
  });
}
const splitLine = `
----------------------------------------------------------`, verifyFlags = {
  ...cli.sharedFlags,
  single: {
    default: !1,
    type: "boolean"
  }
};
function disableCheckText(checkKey) {
  return chalk__default.default.grey(
    outdent__default.default`
              To skip this validation add the following to your package.json:
              "sanityPlugin": {
                 "verifyPackage": {
                    "${checkKey}": false
                 }
              }
          `.trimStart()
  );
}
function createValidator(verifyConfig, flags, errors) {
  return async function(checkKey, task) {
    if (verifyConfig[checkKey] !== !1) {
      const result = await task();
      if (result?.length) {
        result.push(disableCheckText(checkKey));
        const errorMessage = result.join(`

`);
        errors.push(errorMessage), cli.log.error(`
` + errorMessage + splitLine);
      }
    }
    if (flags.single && errors.length)
      throw new Error(
        outdent__default.default`Detected outstanding upgrade issues.

        Fail-fast (--single) mode enabled, stopping validation here.
        `
      );
  };
}
async function runTscMaybe(verifyConfig, ts) {
  if (ts && verifyConfig.tsc !== !1) {
    cli.log.info("All checks ok, running Typescript compiler.");
    const { code } = await runCommand("tsc --build");
    if (code !== 0)
      throw new Error(`Compilation failed. See output above.

` + disableCheckText("tsc"));
  }
}
async function readTSConfig(options) {
  const { basePath, filename } = options, filePath = path__default.default.resolve(basePath, filename);
  if (await _package.fileExists(filePath))
    return await pkgUtils.loadTSConfig({ cwd: basePath, tsconfigPath: filename });
}
exports.createValidator = createValidator;
exports.readTSConfig = readTSConfig;
exports.runTscMaybe = runTscMaybe;
exports.validateImports = validateImports;
exports.verifyFlags = verifyFlags;
//# sourceMappingURL=ts.js.map
